from twisted.application import service
from twisted.internet import reactor, task, defer
from twisted.internet.serialport import SerialPort

from zope.interface import Interface, implements

from agents import IAgentManager, PANTcpAgentManager
from pans import IPANClientFactory
from pans import TanTanPANClientFactory
from uarts import SerialEcho
from utils import loadConfig


class IPANService(Interface):
    """ A client service made to connect Physical-Area-Networks to a
        central server over the network.
    """
    def startAgent():
        """
        """

    def startCOMM():
        """
        """

    def closeAgent():
        """
        """

    def closeCOMM():
        """
        """


class TanTanPANClientService(service.Service):

    implements(IPANService)

    def __init__(self, *args, **kwargs):
        print "PAN CLIENT ARGS; KWARGS: \n{0}\n{1}".format(
                repr(args), repr(kwargs))
        if 'options' in kwargs:
            self.config = kwargs['options']
        else:
            self.config = loadConfig('config.json')
        self.pan = {}
        self.agents = {}
        self.managers = {}
        self.managers['agents'] = IAgentManager(self)

        """
        """
    def startAgent(self, pan_id):
        host = self.config['server']['host']
        port = self.config['server']['port']
        manager = self.managers['agents']
        manager.agents[pan_id] = "CONNECTING"
        client = manager.addAgent(pan_id, host, port)
        return client


    def closeAgent(self, pan_id):
        """
        """

    def startCOMM(self, pan_id):
        agents = self.managers['agents'].agents
        if pan_id in agents:
            agent = agents[pan_id]
        else:
            agent = None
        return self._startCOMM(agent)

    def closeCOMM(self, pan_id):
        """
        """

    def get_pan_info(self, pan_id):
        return defer.succeed(self.config['networks'].get(pan_id, None))

    def getPANs(self):
        return self.config['networks'].keys()

    def _startCOMM(self, agent):

        if agent:
            pan_id = agent.factory.pan_id
        else:
            pan_id = None

        def open_serial_port(agent):
            if pan_id in self.config['networks']:
                pan = self.config['networks'][pan_id]
                port = pan.get('port')
                baud = pan.get('baud')
                print "opening PAN {0}, PORT {1}, BAUD {2}".format(pan_id, port, baud)
                return SerialPort(
                    SerialEcho(agent),
                    port,
                    reactor,
                    baudrate=baud
                    )
            return None

        def connection_error(reason):
            print "Failed to connect", pan_id, reason.value
            if pan_id in self.pan:
                del self.pan[pan_id]
            return None

        def setComm(serial):
            if serial:
                print "Setting COMM", pan_id, serial
                self.pan[pan_id] = serial
                return serial

        d = defer.Deferred()
        d.addCallback(open_serial_port)
        d.addErrback(connection_error)
        d.addCallback(setComm)
        return d.callback(agent)

    def startPANClients(self):
        for pan_id in self.getPANs():
            self.startPANClient(pan_id)

    def startPANClient(self, pan_id):
        client = self.startAgent(pan_id)
        client.addCallback(self._startCOMM)

    def stopPANClients(self):
        for pan_id, client in self.managers['agents'].agents.items():
            client.transport.loseConnection()
            print pan_id, " PAN closed"

    def startService(self):
        self.startPANClients()
        service.Service.startService(self)

    def stopService(self):
        self.stopPANClients()
        service.Service.stopService(self)



from twisted.internet import reactor, task, defer
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.serialport import SerialPort

from zope.interface import Interface, implements
from autobahn.wamp1.protocol import WampClientFactory

from pans import IPANClientFactory
from uarts import SerialEcho
from utils import compDictKeys
from wamp import WAMPClientProtocol


class IAgentManager(Interface):
    """
    """


class PANTcpAgentManager(object):

    implements(IAgentManager)

    def __init__(self, service):
        self.service = service
        self.agents = {}
        self.agent_check = task.LoopingCall(self.checkAgents)
        self.agent_check.start(6)

    def makeFactory(self, pan_id=None):
        factory = IPANClientFactory(self.service)
        factory.pan_id = pan_id
        return factory

    def addAgent(self, pan_id, host, port):
        factory = self.makeFactory(pan_id)
        endpoint = TCP4ClientEndpoint(reactor, host, port)
        agent = endpoint.connect(factory)

        def setConn(client):
            host = client.transport.getHost()
            peer = client.transport.getPeer()
            print "{0} AGENT; {1}; {2}".format(pan_id, host, peer)
            self.agents[pan_id] = client
            return client

        def failedConn(failure):
            print "Could not connect {0} AGENT; {1}".format(pan_id, failure.value)
            if pan_id in self.agents:
                self.agents[pan_id] = None
            return None

        agent.addCallback(setConn)
        agent.addErrback(failedConn)
        return agent

    def hasDisconnectedAgents(self):
        agents = self.agents
        networks = self.service.config.get('networks', {})
        diff = compDictKeys(agents, networks)
        return len(diff) > 0

    def checkAgents(self):
        print "CHECKING AGENTS"
        #if self.hasDisconnectedAgents():
        for pan_id in self.agents:
            agent = self.agents.get(pan_id, None)
            if agent and agent == "CONNECTING":
                print "Agent present; already trying to connect", pan_id
            elif agent:
                print "Agent present", pan_id, agent.connected
            else:
                print "Agent NOT present", pan_id
                self.agents[pan_id] = "CONNECTING"
                self.service.startPANClient(pan_id)


class PANWampAgentManager(PANTcpAgentManager):

    implements(IAgentManager)

    def makeFactory(self, pan_id=None):
        factory = WampClientFactory("ws://localhost:9000", debugWamp = debugW)
        factory.protocol = WAMPClientProtocol
        factory.pan_id = pan_id
        return factory



# -*- coding: utf-8 -*-
import json
from pprint import pprint

from twisted.application import internet, service
from twisted.internet import protocol, reactor, defer, task
from twisted.internet.serialport import SerialPort
from twisted.internet.endpoints import clientFromString
from twisted.internet.endpoints import serverFromString, \
                                       TCP4ClientEndpoint, \
                                       TCP4ServerEndpoint
from twisted.python import components
from twisted.python.log import ILogObserver, FileLogObserver
from twisted.python.logfile import DailyLogFile

from txXBee.protocol import txXBee

from zope.interface import Interface, implements

from autobahn.twisted.websocket import WrappingWebSocketServerFactory, \
                                       WrappingWebSocketClientFactory

from autobahn.wamp1.protocol import WampServerFactory, \
                                    WampServerProtocol, \
                                    WampClientFactory, \
                                    WampClientProtocol, \
                                    exportRpc


debug = False
debugW = False


class WAMPServerProtocol(WampServerProtocol):

    def onSessionOpen(self):
        print "WAMP connection started"
        self.registerForPubSub("http://www.tantan.org/api/sensores#", True)
        self.registerForPubSub("http://api.tantan.net/pan/", True)
        self.call("http://api.tantan.net/pan#getPanId").addCallback(self.onPanId)

    def onPanId(self, result):
        print "WAMP PAN ID", result
        self.dispatch("http://api.tantan.net/pan/id", result)

    def connectionLost(self, reason):
        print "WAMP connection lost"
        WampServerProtocol.connectionLost(self, reason)


class WAMPClientProtocol(WampClientProtocol):

    @exportRpc("getPanId")
    def getPanId(self):
        return self.factory.pan_id

    def onSessionOpen(self):
        pan_id = self.factory.pan_id
        print "WAMP client connected" #, pan_id
        self.registerForRpc(self, "http://api.tantan.net/pan#")

    def connectionLost(self, reason):
        print "WAMP agent connection lost"
        WampClientProtocol.connectionLost(self, reason)



from twisted.internet import protocol
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import ReconnectingClientFactory

from zope.interface import Interface, implements

from zigbee import PANZigBeeProtocol

class IPANServerFactory(Interface):
    """ A factory for clients made to publish at a central node.
    """


class TanTanPANServerFactory(protocol.ServerFactory):

    implements(IPANServerFactory)

    protocol = PANZigBeeProtocol

    def __init__(self, service):
        self.service = service

    def getNetwork(self, pan_id=None):
        return self.service.getPAN(pan_id)

    def getNetworks(self, pan_id=None):
        return self.service.getPANs()


class TanTanPANClientProtocol(Protocol):

    def connectionMade(self):
        pans = self.factory.getNetworkInfo()
        self.factory.resetDelay()
        print "Service PAN IDs", repr(pans)

    def dataReceived(self, data):
        #print "Client received DATA:", data
        pans = self.factory.service.pan
        if self.factory.pan_id in pans:
            pans[self.factory.pan_id].protocol.transport.write(data)

    def connectionLost(self, reason):
        print "AGENT connection LOST", self.factory.pan_id, self.factory.service.pan.keys()
        pans = self.factory.service.pan
        agents = self.factory.service.managers['agents'].agents

        if self.factory.pan_id in agents:
            print "AGENT FACTORY removed" #, dir(self.factory)
            agents[self.factory.pan_id] = None

        if self.factory.pan_id in pans:
            pans[self.factory.pan_id].protocol.transport.loseConnection()


class IPANClientFactory(Interface):
    """ A factory for clients made to publish at a central node.

        The PAN client factory manages connections to Physical-
        Area-Networks (PANs), by controlling UART connections
        via the PAN service.
    """


class TanTanPANClientFactory(ReconnectingClientFactory):

    implements(IPANClientFactory)

    protocol = TanTanPANClientProtocol
    maxDelay = 5

    def __init__(self, service):
        self.service = service
        self.pan_id = None

    def getNetworkInfo(self):
        return self.service.config['networks'].keys()

    #def buildProtocol(self, addr):
    #    print 'Connected.'
    #    print 'Resetting reconnection delay'
    #    self.resetDelay()
    #    prot = TanTanPANClientProtocol()
    #    prot.factory = self
    #    return self

    #def startedConnecting(self, connector):
    #    #self.resetDelay()

    def clientConnectionLost(self, connector, reason):
        print 'Lost connection.  Reason:', reason
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)

    def clientConnectionFailed(self, connector, reason):
        print 'Connection failed. Reason:', reason
        ReconnectingClientFactory.clientConnectionFailed(self, connector,
                                                         reason)



from twisted.internet.protocol import Protocol


class SerialEcho(Protocol):

    def __init__(self, client):
        self.client = client
        self.pan_id = self.client.factory.pan_id

    def connectionMade(self):
        print "Serial connection made"
        print "CLIENT", repr(self.client)

    def connectionLost(self, reason):
        print "Serial port NOT CONNECTED", self.pan_id
        if self.pan_id in self.client.factory.service.pan:
            del self.client.factory.service.pan[self.pan_id]

    def dataReceived(self, data):
        print data.encode('hex')
        self.client.transport.write(data)





# -*- coding: utf-8 -*-
from twisted.application import service
from twisted.internet import reactor, task, defer
from twisted.internet.serialport import SerialPort

from zope.interface import Interface, implements

from txXBee.protocol import txXBee


class PANZigBeeProtocol(txXBee):

    def __init__(self, *args, **kwds):
        super(PANZigBeeProtocol, self).__init__(*args, **kwds)
        self.pan_id = None
        self.lc = task.LoopingCall(self.getSomeData)
        self.lc.start(10.0)
        #self.pan_status = task.LoopingCall(self.obtainPAN)
        #self.pan_status.start(10.0)

    def connectionMade(self):
        print "Connection MADE", self.transport.getPeer()
        self.state = "SILENT"
        self.getPanId()

    def connectionLost(self, reason):
        if self.pan_id and self.factory.service.agents.get(self.pan_id):
            self.factory.service.agents[self.pan_id].sendClose()
            del self.factory.service.agents[self.pan_id]
        if self.pan_id and self.factory.service.networks.get(self.pan_id):
            del self.factory.service.networks[self.pan_id]

    def handle_packet(self, packet):
        print "{0}".format(self._frame.raw_data.encode('hex'))
        def print_debug(packet):
            print "FRAME", self._frame.raw_data.encode('hex')
            print "PACKET", packet

        agent = self.factory.service.agents.get(self.pan_id, None)

        if self.is_PANID(packet):
            self.pan_id = pan_id = packet['parameter'].encode('hex').lstrip('0')
            self.factory.service.networks[pan_id] = self
            self.factory.service.startWAMPClient(pan_id)
            self.state = "PUBLISH"
            print "GOT PAN ID", self.pan_id, self.state
        if self.is_ND(packet):
            #print_debug(packet)
            addr_long = packet['parameter']['source_addr_long'].encode('hex')
            addr = packet['parameter']['source_addr'].encode('hex')
            if agent:
                agent.publish("http://api.tantan.net/pan/sensors#nd", [self.pan_id, addr_long, addr])
        if self.state == "SILENT":
            self.getPanId()
        elif self.state == "PUBLISH":
            if self.is_RX(packet):
                rx = packet['rf_data']
                self.handle_rx(packet, agent)

    def is_target_type(self, packet, field, target):
        if field in packet and packet[field].lower() == target:
            return True
        return False

    def is_AT_RESPONSE(self, packet):
        return self.is_target_type(packet, 'id', 'at_response')

    def is_ND(self, packet):
        is_atresp = self.is_AT_RESPONSE(packet)
        is_nd = self.is_target_type(packet, 'command', 'nd')
        return is_atresp and is_nd

    def is_RX(self, packet):
        return self.is_target_type(packet, 'id', 'rx')

    def is_PANID(self, packet):
        is_atresp = self.is_AT_RESPONSE(packet)
        is_nd = self.is_target_type(packet, 'command', 'id')
        return is_atresp and is_nd

    def get_zb_node_info(self, packet):
        try:
            resp = {}
            resp['id'] = packet["source_addr_long"].encode('hex')
            resp['laddr'] = packet["source_addr_long"].encode('hex')
            resp['addr'] = packet["source_addr"].encode('hex')
            resp['data'] = packet["rf_data"] or ''
            return resp
        except:
            return None

    def handle_rx(self, packet, agent):
        if agent:
            resp = {}
            #resp['name'] = ZB_reverse.get(packet["source_addr_long"], "Unknown")
            resp = self.get_zb_node_info(packet)
            rout = "RX:"
            for (key, item) in resp.items():
                rout += "{0}-{1}:".format(key, item)
            msg = "{0}:{1}:RX:".format(resp['id'], resp['addr'], resp['data']) + repr(resp['data'])
            #print 'Evt id: {0}\nVal: {1}'.format(str(resp['name']), resp['data'].decode('utf8'))
            evt = {'id': resp['id'],
                    'type': 'rx',
                    'data': resp['data'],
                    }
            node_id = resp['id']
            data = resp['data']
            data_lines = data.splitlines()
            #print "VAL LINES", val_lines
            for l in data_lines:
                uri = "#".join(["http://www.tantan.org/api/sensores", node_id])
                agent.publish(uri, {'node_id': node_id, 'msg': l})

    def getPanId(self):
        reactor.callFromThread(self.send,
                "at",
                frame_id="\x02",
                command="ID"
                )

    def getC0ID(self, dest_addr_long, dest_addr, txdata):
        reactor.callFromThread(self.send,
                "at",
                command="ID",
                )

    def obtainPAN(self):
        if self.pan_id and self.status == "SILENT":
            self.status == "PUBLISH"
        elif self.status == "SILENT":
            self.getPanId()

    def getSomeData(self):
        self.sendND()
        self.allTX('HOLA')
        self.nodeTX(
                dest_addr_long="\x00\x13\xa2\x00\x40\xa4\xd4\x93",
                dest_addr="\xa7\xe8",
                txdata=u"Joaquín".encode('utf8')
                )

    def sendND(self):
        reactor.callFromThread(self.send,
                "remote_at",
                frame_id="\x01",
                command="ND",
                )

    def nodeTX(self, dest_addr_long, dest_addr, txdata):
        reactor.callFromThread(self.send,
                "tx",
                frame_id="\x03",
                dest_addr_long=dest_addr_long,
                dest_addr=dest_addr,
                data=txdata,
                )
    def allTX(self, txdata):
        reactor.callFromThread(self.send,
                "tx",
                frame_id="\x03",
                dest_addr_long="\x00\x00\x00\x00\x00\x00\xff\xff",
                dest_addr="\xff\xfe",
                data=txdata,
                )





# -*- coding: utf-8 -*-
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
    
setup(
    name="TanTan",
    version="0.1.1",
    author="Joaquin Rosales",
    author_email="globojorro@gmail.com",
    packages=["tantan"],
    scripts=[],
    url="https://pypi.python.org/pypi/TanTan",
    license="LICENSE.txt",
    description="An open-source application for communicating with \
        Physical-Area-Networks using the ZigBee protocol",
    long_description="""A collection of services, protocols, clients and
        servers based on the Twisted framework.
        """,
    install_requires=[
        "setuptools >= 3.4.4",
        "pySerial",
        "pyOpenSSl",
        "Twisted == 13.1.0",
        "autobahn == 0.8.6",
        "paisley",
        "txXBee"
    ],
)



import json
from pprint import pprint

def handle_factory_db(factory, response, topicUrl):
    pprint(response, depth=2)
    evt = response
    factory.dispatch(topicUrl, evt) 

def loadConfig(filepath):
    try:
        cfg = json.load(open(filepath, 'r'))
        pprint(cfg, depth=4)
    except:
        cfg = {}
    return cfg

def saveConfig(cfg, filepath):
    try:
        json.dump(cfg, open(filepath, 'w'), indent=4)
        pprint(cfg, depth=4)
    except:
        print "Could not save config"

def compDictKeys(a, b):
    set_a = set(a.keys())
    set_b = set(b.keys())
    return set_b.difference(set_a)
